import subprocess
from pygeoweaver.utils import get_geoweaver_jar_path, get_java_bin_path, get_root_dir


def helpwith(
    command_list: list = [],
):
    target_cmd_args = [get_java_bin_path(), "-jar", get_geoweaver_jar_path()]
    if len(command_list) > 0:
        for i in range(len(command_list) - 1):
            target_cmd_args.append(command_list[i])
        target_cmd_args.append("help")
        target_cmd_args.append(command_list[-1])
    else:
        target_cmd_args.append("help")
    subprocess.run(target_cmd_args, cwd=f"{get_root_dir()}/")



import requests


class BaseAPI:
    def __init__(self, base_url):
        self.base_url = base_url

    def _call_api(self, endpoint, method='GET', data=None):
        url = self.base_url + endpoint
        headers = {
            'Content-Type': 'application/json'
        }

        try:
            if method.upper() == 'POST':
                response = requests.post(url, headers=headers, data=json.dumps(data))
            elif method.upper() == 'GET':
                response = requests.get(url, headers=headers, params=data)
            elif method.upper() == 'PUT':
                response = requests.put(url, headers=headers, data=json.dumps(data))
            elif method.upper() == 'DELETE':
                response = requests.delete(url, headers=headers, data=json.dumps(data))
            else:
                raise ValueError(f"Unsupported HTTP method: {method}")

            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")
        except Exception as err:
            print(f"An error occurred: {err}")



import inspect
import io
import logging
from functools import wraps
import os
import sys

from pygeoweaver.runtime_tags.pgw_process import geoweaver_context


logger = logging.getLogger(__name__)


def pygeoweaver_workflow(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        workflow_name = func.__name__
        
        # Set the current workflow in context
        geoweaver_context['current_workflow'] = workflow_name
        
        # Execute the workflow function
        result = func(*args, **kwargs)
        
        # Clear the current workflow in context
        geoweaver_context['current_workflow'] = None
        
        # Retrieve the process calls for this workflow
        process_calls = geoweaver_context['process_calls'].get(workflow_name, [])
        
        # Print process calls (for demonstration)
        for call in process_calls:
            print(f"Process Name: {call['name']}")
            print("Source Code:")
            print(call['code'])
            print("Logs:")
            print(call['log'])
            print("-" * 40)
        
        return result
    
    return wrapper




import getpass
import json
import os
import subprocess

import requests

from pygeoweaver.constants import GEOWEAVER_DEFAULT_ENDPOINT_URL
from pygeoweaver.utils import (
    download_geoweaver_jar,
    get_geoweaver_jar_path,
    get_java_bin_path,
    get_root_dir,
    get_spinner,
)
from halo import Halo


def run_process(
    *,
    process_id: str,
    host_id: str,
    password: str = None,
    environment: str = None,
    sync_path: os.PathLike = None,
):
    """
    Run a process

    Args: process_id - required
        host_id - required
        password - optional
        environment - optional
    """
    with get_spinner(text=f'Running Geoweaver process {process_id}...', spinner='dots'):
        if password is None:
            # prompt to ask for password
            password = getpass.getpass(f"Enter password for host - {host_id}: ")

        if sync_path:
            if not os.path.exists(sync_path):
                print("The specified path does nto exists")
            print("Updating code on workflow with the given file path.\n")
            f = open(sync_path, "r")
            context = f.read()
            f.close()
            details = requests.post(
                f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/detail",
                data={"type": "process", "id": process_id},
            ).json()
            details["code"] = context
            requests.post(
                f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/edit/process",
                data=json.dumps(details),
                headers={"Content-Type": "application/json"},
            )

        download_geoweaver_jar()
        subprocess.run(
            [
                get_java_bin_path(),
                "-jar",
                get_geoweaver_jar_path(),
                "run",
                "process",
                f"--host={host_id}",
                f"--password={password}",
                f"--environment={environment}",
                process_id,
            ],
            cwd=f"{get_root_dir()}/",
        )


def run_workflow(
    *,
    workflow_id: str,
    workflow_folder_path: str = None,
    workflow_zip_file_path: str = None,
    environment_list: str = None,
    host_list: str = None,
    password_list: str = None,
    sync_path: os.PathLike = None,
):
    """
    Usage: <main class> run workflow [-d=<workflowFolderPath>]
                                    [-f=<workflowZipPath>] [-e=<envs>]...
                                    [-h=<hostStrings>]... [-p=<passes>]...
                                    <workflowId>
        <workflowId>           workflow id to run
    -d, --workflow-folder-path=<workflowFolderPath>
                                geoweaver workflow folder path
    -e, --environments=<envs>  environments to run on. List of environment ids with comma as separator
    -f, --workflow-zip-file-path=<workflowZipPath>
                                workflow package or path to workflow zip to run
    -h, --hosts=<hostStrings>  hosts to run on. list of host ids with comma as separator.
    -p, --passwords=<passes>   passwords to the target hosts. list of passwords with comma as separator.
    """
    with get_spinner(text=f'Running Geoweaver workflow {workflow_id}...', spinner='dots'):
        download_geoweaver_jar()

        if password_list is None:
            # prompt to ask for password
            password_list = []
            for host in host_list.split(","):
                password = getpass.getpass(f"Enter password for host - {host}: ")
                password_list.append(password)
        elif len(password_list.split(",")) != len(host_list.split(",")):
            raise RuntimeError("The password list length doesn't match host list")

        password_list = ",".join(password_list)
        if sync_path:
            from . import sync_workflow

            sync_workflow(workflow_id=workflow_id, sync_to_path=sync_path)

        if not workflow_id and not workflow_folder_path and not workflow_zip_file_path:
            raise RuntimeError(
                "Please provide at least one of the three options: workflow id, "
                "folder path or zip path"
            )

        if workflow_id and not workflow_folder_path and not workflow_zip_file_path:
            command = [
                get_java_bin_path(),
                "-jar",
                get_geoweaver_jar_path(),
                "run",
                "workflow",
                workflow_id,
                "-h",
                host_list,
                "-p",
                password_list,
            ]
            if environment_list:
                command.extend(["-e", environment_list])
            subprocess.run(command, cwd=f"{get_root_dir()}/")

        if workflow_folder_path and not workflow_zip_file_path:
            # command to run workflow from folder
            command = [
                get_java_bin_path(),
                "-jar",
                get_geoweaver_jar_path(),
                "run",
                "workflow",
                workflow_id,
                "-d",
                workflow_folder_path,
                "-h",
                host_list,
                "-p",
                password_list
            ]
            if environment_list:
                command.extend(["-e", environment_list])
            subprocess.run(command, cwd=f"{get_root_dir()}/")

        if not workflow_folder_path and workflow_zip_file_path:
            command = [
                get_java_bin_path(),
                "-jar",
                get_geoweaver_jar_path(),
                "run",
                "workflow",
                workflow_id,
                "-f",
                workflow_zip_file_path,
                "-h",
                host_list,
                "-p",
                password_list,
            ]
            if environment_list:
                command.extend(["-e", environment_list])
            subprocess.run(command, cwd=f"{get_root_dir()}/")



"""
The main function of pygeoweaver
To run in CLI mode. 
"""
import logging
import os
import typing
import click
from pygeoweaver import (
    detail_host,
    detail_process,
    detail_workflow,
    export_workflow,
    show_history,
    import_workflow,
    list_hosts,
    list_processes,
    list_workflows,
    start,
    stop,
    reset_password,
    run_process,
    run_workflow,
    helpwith,
    clean_h2db,
)
from pygeoweaver.commands.pgw_cleanup import cleanup_workspace
from pygeoweaver.constants import GEOWEAVER_DEFAULT_ENDPOINT_URL
from pygeoweaver.commands.pgw_create import create_process, create_process_from_file, create_workflow
from pygeoweaver.commands.pgw_detail import get_process_code
from pygeoweaver.commands.pgw_find import get_process_by_id, get_process_by_language, get_process_by_name
from pygeoweaver.commands.pgw_history import get_process_history, get_workflow_history
from pygeoweaver.commands.pgw_list import list_processes_in_workflow
from pygeoweaver.commands.pgw_sync import sync, sync_workflow
from pygeoweaver.commands.pgw_upgrade import upgrade_geoweaver
from pygeoweaver.pgw_log_config import setup_logging
from pygeoweaver.server import check_geoweaver_status, show
from halo import Halo
from pygeoweaver.utils import get_spinner


setup_logging()


@click.group()
def geoweaver():
    """
    Geoweaver CLI: A tool for managing workflows and processes.
    """
    pass


@geoweaver.command("start")
@click.option('--force-download', is_flag=True, help='Force overwrite the Geoweaver JAR file if it already exists.')
@click.option('--force-restart', is_flag=True, help='Force to restart Geoweaver even when it is already running.')
def start_command(force_download, force_restart):
    """
    Start the Geoweaver application.
    """
    start(force_download=force_download, force_restart=force_restart, exit_on_finish=True)


@geoweaver.command("stop")
def stop_command():
    """
    Start the Geoweaver application.
    """
    stop(exit_on_finish=True)

@geoweaver.command("show")
@click.option('--geoweaver-url', default=GEOWEAVER_DEFAULT_ENDPOINT_URL, help='Geoweaver URL (default is GEOWEAVER_DEFAULT_ENDPOINT_URL)')
def show_command(geoweaver_url):
    """
    Show graphical user interface of Geoweaver in browser

    Args:
        geoweaver_url (_type_): _description_
    """
    show(geoweaver_url)


@geoweaver.command("reset_password")
def reset_password_command():
    """
    Reset the password for localhost.
    """
    reset_password()


@geoweaver.command("cleanup_workspace")
def cleanup_workspace_command():
    """
    Clean up temporary directories in the gw-workspace folder.
    """
    cleanup_workspace()


@geoweaver.group("create")
def create_command():
    """
    Create commands for Geoweaver.
    """
    pass


@create_command.command("process")
@click.option('--lang', prompt='Programming Language', help='The programming language of the process')
@click.option('--description', prompt='Description', help='The description of the process')
@click.option('--name', prompt='Name', help='The name of the process')
@click.option('--code', prompt='Code', help='The code of the process')
@click.option('--file-path', prompt='File Path', type=click.Path(exists=True), help='The path to the file containing the code')
@click.option('--owner', default='111111', help='The owner of the process (default: "111111")')
@click.option('--confidential', is_flag=True, help='Set if the process is confidential')
def create_process_command(lang, description, name, code, file_path, owner, confidential):
    """
    Create a process with given code or given file_path, which ever is valid first.
    """
    if code is not None:
        create_process(lang, description, name, code, owner, confidential)
    elif file_path is not None:
        create_process_from_file(lang, description, name, file_path, owner, confidential)


@create_command.command("workflow")
@click.option('--description', prompt='Description', help='The description of the workflow')
@click.option('--edges', prompt='Edges', help='The edges of the workflow')
@click.option('--name', prompt='Name', help='The name of the workflow')
@click.option('--nodes', prompt='Nodes', help='The nodes of the workflow')
@click.option('--owner', default='111111', help='The owner of the workflow (default: "111111")')
@click.option('--confidential', is_flag=True, help='Set if the workflow is confidential')
def create_workflow_command(description, edges, name, nodes, owner, confidential):
    """
    Create a workflow with given data if valid.
    """
    create_workflow(description, edges, name, nodes, owner, confidential)


@geoweaver.group("detail")
def detail_command():
    """
    Detail commands for Geoweaver.
    """
    pass


@detail_command.command("workflow")
@click.argument('workflow_id', type=str)
def detail_workflow_command(workflow_id):
    """
    Display detailed information about a workflow.

    :param workflow_id: The ID of the workflow.
    :type workflow_id: str
    """
    detail_workflow(workflow_id)


@detail_command.command("process")
@click.argument('process_id', type=str)
def detail_process_command(process_id):
    """
    Display detailed information about a process.

    :param process_id: The ID of the process.
    :type process_id: str
    """
    detail_process(process_id)


@detail_command.command("host")
@click.argument('host_id', type=str)
def detail_host_command(host_id):
    """
    Display detailed information about a host.

    :param host_id: The ID of the host.
    :type host_id: str
    """
    detail_host(host_id)


@detail_command.command("code")
@click.argument('process_id', type=str)
def get_process_code_command(process_id):
    """
    Get the code of a process.

    :param process_id: The ID of the process.
    :type process_id: str
    """
    get_process_code(process_id)


@geoweaver.group("export")
def export_command():
    """
    Export commands for Geoweaver.
    """
    pass


@export_command.command("workflow")
@click.argument('workflow_id', type=str)
@click.option('--mode', default=4, type=int, help='Exportation mode options: \n1 - workflow only, 2 - workflow with process code, 3 - workflow with process code and only good history, 4 - workflow with process code and all the history. Default option is 4.')
@click.argument('target_file_path', type=click.Path())
@click.option('--unzip', is_flag=True, help='Unzip the exported file.')
@click.option('--unzip-directory-name', help='Specify the directory name when unzipping.')
def export_workflow_command(workflow_id, mode, target_file_path, unzip, unzip_directory_name):
    """
    Export a Geoweaver workflow.

    :param workflow_id: Geoweaver workflow ID.
    :type workflow_id: str
    :param mode: Exportation mode options: 1 - workflow only, 2 - workflow with process code, 3 - workflow with process code and only good history, 4 - workflow with process code and all the history. Default option is 4.
    :type mode: int
    :param target_file_path: Target file path to save the workflow zip.
    :type target_file_path: str
    :param unzip: Unzip the exported file.
    :type unzip: bool
    :param unzip_directory_name: Specify the directory name when unzipping.
    :type unzip_directory_name: str
    """
    export_workflow(workflow_id, mode, target_file_path, unzip, unzip_directory_name)


@geoweaver.group("find")
def find_command():
    """
    Find commands for Geoweaver.
    """
    pass


@find_command.command("name")
@click.argument('process_name', type=str)
def get_process_by_name_command(process_name):
    """
    Get processes by their name.

    :param process_name: The name of the process.
    :type process_name: str
    """
    return get_process_by_name(process_name)


@find_command.command("id")
@click.argument('process_id', type=str)
def get_process_by_id_command(process_id):
    """
    Get a process by its ID.

    :param process_id: The ID of the process.
    :type process_id: str
    """
    return get_process_by_id(process_id)
    
    
@find_command.command("language")
@click.argument('language', type=str)
def get_process_by_language_command(language):
    """
    Get processes by their programming language.

    :param language: The programming language of the processes.
    :type language: str
    """
    return get_process_by_language(language)

@geoweaver.group("history")
def history_command():
    """
    history commands for Geoweaver.
    """
    pass

@history_command.command("show")
@click.argument('history_id', type=str)
def show_history_command(history_id):
    """
    Show history for a workflow or process.

    :param history_id: The ID of the history.
    :type history_id: str
    """
    show_history(history_id)
    
    
@history_command.command("get_process")
@click.argument('process_id', type=str)
def get_process_history_command(process_id):
    """
    Get list of history for a process using process id.

    :param process_id: The ID of the process.
    :type process_id: str
    """
    get_process_history(process_id)


@history_command.command("get_workflow")
@click.argument('workflow_id', type=str)
def get_workflow_history_command(workflow_id):
    """
    Get list of history for a workflow using workflow id.

    :param workflow_id: The ID of the workflow.
    :type workflow_id: str
    """
    get_workflow_history(workflow_id)


@geoweaver.group("import")
def import_command():
    """
    import commands for Geoweaver.
    """
    pass

@import_command.command("workflow")
@click.argument('workflow_zip_file_path', type=click.Path(exists=True, dir_okay=False))
def import_workflow_command(workflow_zip_file_path):
    """
    Import a workflow from a zip file.

    :param workflow_zip_file_path: The path to the Geoweaver workflow zip file.
    :type workflow_zip_file_path: str
    """
    import_workflow(workflow_zip_file_path)


@geoweaver.group("list")
def list_command():
    """
    list commands for Geoweaver.
    """
    pass


@list_command.command("host")
def list_hosts_command():
    """
    List all hosts in Geoweaver.

    Downloads the Geoweaver JAR, then runs the 'list' command with the '--host' option.

    Note: Requires Geoweaver to be initialized and the JAR file to be available.
    """
    list_hosts()


@list_command.command("process")
def list_processes_command():
    """
    List all processes in Geoweaver.

    Downloads the Geoweaver JAR, ensures executable permissions, and then runs the 'list' command with the '--process' option.

    Note: Requires Geoweaver to be initialized and the JAR file to be available.
    """
    list_processes()


@list_command.command()
@click.argument('workflow_id', type=str)
def list_processes_in_workflow_command(workflow_id):
    """
    List processes in a specific workflow.

    Downloads the Geoweaver JAR and queries the Geoweaver server for details of a workflow.
    Extracts information about nodes in the workflow and returns a list of dictionaries containing 'title' and 'id'.

    :param workflow_id: The ID of the workflow.
    :type workflow_id: str
    :return: List of dictionaries containing 'title' and 'id' of processes in the workflow.
    :rtype: list
    """
    list_processes_in_workflow(workflow_id)


@list_command.command("workflow")
def list_workflows_command():
    """
    List all workflows in Geoweaver.

    Downloads the Geoweaver JAR, then runs the 'list' command with the '--workflow' option.

    Note: Requires Geoweaver to be initialized and the JAR file to be available.
    """
    list_workflows()


@geoweaver.group("run")
def run_command():
    """
    Run commands for Geoweaver.
    """
    pass


@run_command.command()
@click.option('--process-id', required=True, type=str, help='ID of the process.')
@click.option('--host-id', required=True, type=str, help='ID of the host.')
@click.option('--password', type=str, help='Password for authentication.')
@click.option('--environment', type=str, help='Environment for the process.')
@click.option('--sync-path', type=click.Path(exists=True), help='Path for synchronization.')
def run_process_command(process_id, host_id, password, environment, sync_path):
    """
    Run a process.

    Args:
        --process-id: ID of the process. (required)
        --host-id: ID of the host. (required)
        --password: Password for authentication. (optional)
        --environment: Environment for the process. (optional)
        --sync-path: Path for synchronization. (optional)
    """
    run_process(
        process_id=process_id,
        host_id=host_id,
        password=password,
        environment=environment,
        sync_path=sync_path,
    )


@run_command.command()
@click.argument('workflow_id', type=str)
@click.option('--workflow-folder-path', '-d', type=str, help='Geoweaver workflow folder path.')
@click.option('--workflow-zip-file-path', '-f', type=str, help='Path to workflow zip file.')
@click.option('--environments', '-e', type=str, help='Environments to run on. List of environment ids with comma as separator.')
@click.option('--hosts', '-h', type=str, help='Hosts to run on. List of host ids with comma as separator.')
@click.option('--passwords', '-p', type=str, help='Passwords to the target hosts. List of passwords with comma as separator.')
@click.option('--sync-path', type=click.Path(exists=True), help='Path for synchronization.')
def run_workflow_command(workflow_id, workflow_folder_path, workflow_zip_file_path, environments, hosts, passwords, sync_path):
    """
    Run a workflow.

    Args:
        <workflow_id>: Workflow ID to run.
        -d, --workflow-folder-path: Geoweaver workflow folder path. (optional)
        -f, --workflow-zip-file-path: Path to workflow zip file. (optional)
        -e, --environments: Environments to run on. List of environment ids with comma as separator. (optional)
        -h, --hosts: Hosts to run on. List of host ids with comma as separator. (optional)
        -p, --passwords: Passwords to the target hosts. List of passwords with comma as separator. (optional)
    """
    run_workflow(
        workflow_id=workflow_id,
        workflow_folder_path=workflow_folder_path,
        workflow_zip_file_path=workflow_zip_file_path,
        environment_list=environments,
        host_list=hosts,
        password_list=passwords,
        sync_path=sync_path,
    )


@geoweaver.group("sync")
def sync_command():
    """
    sync commands for Geoweaver.
    """
    pass

@sync_command.command("process")
@click.option('--process-id', required=True, type=str, help='The ID of the Geoweaver process.')
@click.option('--local-path', required=True, type=click.Path(exists=True, file_okay=True, dir_okay=True),
              help='The local path to save or load the process code.')
@click.option('--direction', required=True, type=click.Choice(['download', 'upload'], case_sensitive=False),
              help='The direction of the sync, either "download" or "upload".')
def sync_process_command(process_id: str, local_path: typing.Union[str, os.PathLike], direction: str):
    """
    Sync code for a Geoweaver process between the local machine and the Geoweaver server.

    :param process_id: The ID of the Geoweaver process.
    :type process_id: str
    :param local_path: The local path to save or load the process code.
    :type local_path: Union[str, os.PathLike]
    :param direction: The direction of the sync, either "download" or "upload".
    :type direction: str
    """
    
    sync(process_id, local_path, direction)
    

@sync_command.command("workflow")
@click.option('--workflow-id', required=True, type=str, help='The ID of the Geoweaver workflow.')
@click.option('--sync-to-path', required=True, type=click.Path(exists=True, file_okay=False, dir_okay=True),
              help='The local path to sync the Geoweaver workflow.')
def sync_workflow_command(workflow_id: str, sync_to_path: typing.Union[str, os.PathLike]):
    """
    Sync a Geoweaver workflow, including its code and history, between the local machine and the Geoweaver server.

    :param workflow_id: The ID of the Geoweaver workflow.
    :type workflow_id: str
    :param sync_to_path: The local path to sync the Geoweaver workflow.
    :type sync_to_path: Union[str, os.PathLike]
    """
    
    sync_workflow(workflow_id, sync_to_path)


@geoweaver.command("status")
def status():
    """
    Check the status of Geoweaver.
    """
    with get_spinner(text='Checking Geoweaver status...', spinner='dots'):
        geoweaver_running = check_geoweaver_status()
    
    if geoweaver_running:
        click.echo(click.style("Geoweaver is running", fg='green', bold=True))
    else:
        click.echo(click.style("Geoweaver is not running", fg='red', bold=True))


@geoweaver.command("cleanh2db")
@click.option('--h2-jar-path', type=click.Path(exists=True), help='Path to the H2 database JAR file. If not provided, will use h2-2.2.224.jar in the current directory.')
@click.option('--temp-dir', type=click.Path(), help='Path to a temporary directory for the recovery process. If not provided, will create one.')
@click.option('--db-path', type=click.Path(), help='Path to the H2 database files. If not provided, will use ~/h2_hopper_amd_1/gw.')
@click.option('--username', default="geoweaver", help='Username for the H2 database. Defaults to "geoweaver".')
@click.option('--password', help='Password for the H2 database. If not provided, will prompt the user.')
def cleanh2db_command(h2_jar_path, temp_dir, db_path, username, password):
    """
    Clean and reduce the size of the H2 database used by Geoweaver.
    
    This command follows these steps:
    1. Stop Geoweaver if it's running
    2. Create a temporary directory if one is not provided
    3. Copy database files to the temporary directory
    4. Export data from the database to a SQL file
    5. Remove the original database files
    6. Import the SQL file into a new database
    7. Start Geoweaver
    """
    success = clean_h2db(
        h2_jar_path=h2_jar_path,
        temp_dir=temp_dir,
        db_path=db_path,
        username=username,
        password=password
    )
    
    if success:
        click.echo(click.style("H2 database cleanup completed successfully!", fg='green', bold=True))
    else:
        click.echo(click.style("H2 database cleanup failed. Check the logs for details.", fg='red', bold=True))


@geoweaver.command("upgrade")
@click.option('--force', is_flag=True, help='Skip confirmation prompt and force upgrade')
@click.option('--no-start', is_flag=True, help='Do not start Geoweaver after upgrade')
def upgrade_command(force, no_start):
    """
    Upgrade Geoweaver by downloading the latest JAR file.
    
    This command follows these steps:
    1. Stop Geoweaver if it's running
    2. Force download the latest Geoweaver JAR file
    3. Start Geoweaver (unless --no-start is specified)
    """
    success = upgrade_geoweaver(force=force, no_start=no_start)
    
    if success:
        click.echo(click.style("Geoweaver upgrade completed successfully!", fg='green', bold=True))
    else:
        click.echo(click.style("Geoweaver upgrade failed. Check the logs for details.", fg='red', bold=True))


from pygeoweaver.commands.pgw_cleanup import cleanup_workspace

@geoweaver.command("installjdk")
@click.option('--install-dir', help='Custom installation directory for JDK', type=str)
@click.option('--version', default="11.0.18-10", help='JDK version to install', type=str)
def install_jdk_command(install_dir, version):
    """Force install JDK in the specified directory or home directory."""
    from pygeoweaver.jdk_utils import install_jdk_macos, install_jdk_linux, install_jdk_windows
    import platform

    system = platform.system()
    architecture = platform.machine()
    
    if install_dir:
        os.environ['HOME'] = install_dir

    if system == "Darwin":
        if architecture == "x86_64":
            install_jdk_macos(version, "jdk_x64_mac_hotspot")
        elif architecture == "arm64":
            install_jdk_macos(version, "jdk_aarch64_mac_hotspot")
        else:
            click.echo("Unsupported architecture.")

    elif system == "Linux":
        if architecture == "x86_64":
            install_jdk_linux(version, "jdk_x64_linux_hotspot")
        elif architecture == "aarch64":
            install_jdk_linux(version, "jdk_aarch64_linux_hotspot")
        else:
            click.echo("Unsupported architecture.")

    elif system == "Windows":
        if architecture == "AMD64" or architecture == "x86_64":
            install_jdk_windows(version, "jdk_x64_windows_hotspot")
        elif architecture == "x86-32":
            install_jdk_windows(version, "jdk_x86-32_windows_hotspot")
        else:
            click.echo("Unsupported architecture.")

    else:
        click.echo("Unsupported platform.")


if __name__ == "__main__":
    geoweaver()



import os
import json
import shutil
import zipfile

import requests
import typing

from pygeoweaver.constants import *
from pygeoweaver.utils import (
    download_geoweaver_jar,
    copy_files,
    get_spinner,
    safe_exit,
)
from halo import Halo


def overwrite_files(source_dir, destination_dir):
    """
    Overwrite all files from the source directory to the destination directory.
    
    Args:
        source_dir (str): The source directory path.
        destination_dir (str): The destination directory path.
    """
    if not os.path.exists(source_dir):
        print(f"Source directory {source_dir} does not exist.")
        return

    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)
    
    for item in os.listdir(source_dir):
        source_path = os.path.join(source_dir, item)
        destination_path = os.path.join(destination_dir, item)
        
        if os.path.isfile(source_path):
            shutil.copy2(source_path, destination_path)
        elif os.path.isdir(source_path):
            if os.path.exists(destination_path):
                shutil.rmtree(destination_path)
            shutil.copytree(source_path, destination_path)


def sync(process_id: str, local_path: typing.Union[str, os.PathLike], direction: str):
    """
    Sync code for a Geoweaver process between the local machine and the Geoweaver server.

    :param process_id: The ID of the Geoweaver process.
    :param local_path: The local path to save or load the process code.
    :param direction: The direction of the sync, either "download" or "upload".
    """
    with get_spinner(text=f'Sync Geoweaver process {process_id} from database to local folder {local_path}...', 
              spinner='dots'):
        print(f"Proceeding with {direction}\n")
        if direction == "download":
            if not local_path:
                raise Exception("Sync path not found.")
            r = requests.post(
                f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/detail",
                data={"type": "process", "id": process_id},
            ).json()
            code = r["code"]
            decoded_string = code
            file_name = r["name"]
            ext = None
            if r["lang"] == "python":
                ext = ".py"
            elif r["lang"] == "shell":
                ext = ".sh"
            elif r["lang"] == "jupyter":
                ext = "ipynb"
            else:
                raise Exception("Unknown file format.")
            with open(os.path.join(local_path, file_name + ext), "w") as file:
                file.write(decoded_string)
            print(f"Wrote file {file_name + ext} to {local_path}")
        elif direction == "upload":
            if not local_path:
                raise Exception("Sync path not found.")
            process_prev_state = requests.post(
                f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/detail",
                data={"type": "process", "id": process_id},
            ).json()
            with open(local_path, "r") as f:
                f_content = f.read()
                process_prev_state["code"] = f_content
                response = requests.post(
                    f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/edit/process",
                    data=json.dumps(process_prev_state),
                    headers={"Content-Type": "application/json"},
                )
                if response.status_code == 200:
                    print("Process update was successful")
                else:
                    print("Update failed with status code:", response.status_code)
        else:
            raise Exception(
                "Please specify the direction to sync. Choices - [UPLOAD, DOWNLOAD]"
            )

def sync_workflow(workflow_id: str, sync_to_path: typing.Union[str, os.PathLike]):
    """
    Sync a Geoweaver workflow, including its code and history, between the local machine and the Geoweaver server.

    :param workflow_id: The ID of the Geoweaver workflow.
    :param sync_to_path: The local path to sync the Geoweaver workflow.
    """
    with get_spinner(text=f'Sync Geoweaver workflow {workflow_id} from database to local folder {sync_to_path}...', 
              spinner='dots'):
        download_geoweaver_jar()
        # download workflow
        r = requests.post(
            f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/downloadworkflow",
            data=f"id={workflow_id}&option=workflowwithprocesscodeallhistory",
            headers={
                'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'
            }
        )
        if r.status_code != 200:
            print("Error: Fail to prepare the workflow folder.")
            safe_exit(1)
        home_dir = os.path.expanduser("~")
        source_folder = os.path.join(home_dir, "gw-workspace", "temp", workflow_id)
        # check if target workflow path and the unzipped workflow match
        if not sync_to_path:
            raise Exception(
                "Please provide path to workflow that you wish to sync code and history"
            )
        
        target_workflow_json_path = os.path.join(sync_to_path, "workflow.json")
        if os.path.exists(target_workflow_json_path):
            sync_id = json.loads(
                open(target_workflow_json_path, "r").read()
            ).get("id")

            if workflow_id != sync_id:
                print("Error: Workflow ID mismatch, please check the existing workflow.json in the `sync_to_path` path.")
                safe_exit(1)
        copy_files(
            source_folder=source_folder, destination_folder=sync_to_path
        )
        print("Sync is complete.")



import inspect
import logging
from logging.config import fileConfig
import os


def setup_logging():
    log_file = '~/geoweaver.log'
    log_file = os.path.expanduser(log_file)

    # Ensure the directory for the log file exists, create if not
    log_dir = os.path.dirname(log_file)
    os.makedirs(log_dir, exist_ok=True)
    # Get the absolute path of the current script or notebook file
    current_file = inspect.getfile(inspect.currentframe())
    current_file_path = os.path.abspath(current_file)
    current_folder = os.path.dirname(current_file_path)
    # Get the absolute path to the pgw_logging.ini file
    logging_ini_path = os.path.abspath(f'{current_folder}/pgw_logging.ini')

    # Open the pgw_logging.ini file
    with open(logging_ini_path, 'rt') as f:
        config_str = f.read()
        config_str = config_str.replace('%(log_file)s', os.path.abspath(log_file).replace('\\', '/'))

    config_file = f'{current_folder}/logging_temp.ini'
    with open(config_file, 'wt') as f:
        f.write(config_str)

    fileConfig(config_file)
    os.remove(config_file)


def get_logger(class_name):
    """
    Get a logger with the specified class name.
    """
    logger = logging.getLogger(class_name)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    return logger



"""
The main function of pygeoweaver
To run in CLI mode. 
"""
import logging
import time
from unittest.mock import patch
import requests
from pygeoweaver import start, stop

import unittest

from pygeoweaver.constants import GEOWEAVER_DEFAULT_ENDPOINT_URL
from pygeoweaver.pgw_log_config import get_logger
from pygeoweaver.server import show
import pytest


logger = get_logger(__name__)

def test_server_start_stop():
    # Start the server
    start(exit_on_finish=False)

    # Check if the server is up by making a GET request
    response = requests.get(GEOWEAVER_DEFAULT_ENDPOINT_URL)
    assert response.status_code == 200, f"Failed to access URL: {GEOWEAVER_DEFAULT_ENDPOINT_URL}"

    # Stop the server
    stop(exit_on_finish=False)
    
    time.sleep(5)

    # Check that the server has stopped by expecting a connection error
    with pytest.raises(requests.exceptions.ConnectionError):
        print(f"Test {GEOWEAVER_DEFAULT_ENDPOINT_URL}")
        response = requests.get(GEOWEAVER_DEFAULT_ENDPOINT_URL)
        print(response)

    # Stop the server again should have no issue
    stop(exit_on_finish=False)

def test_windows():
    with patch("pygeoweaver.server.check_os") as mock_checkos:
        mock_checkos.return_value = 3
        
        # Check that FileNotFoundError is raised when starting the server
        with pytest.raises(FileNotFoundError):
            start(exit_on_finish=False)
        
        # Check that FileNotFoundError is raised when stopping the server
        with pytest.raises(FileNotFoundError):
            stop(exit_on_finish=False)

def test_show_gui():
    with patch("pygeoweaver.webbrowser.open") as mock_browser_open:
        show()
        mock_browser_open.assert_called_once()

        with patch("pygeoweaver.server.check_ipython") as mock_checkipython:
            mock_checkipython.return_value = True
            show()
            mock_browser_open.assert_called_once()




from pydantic import BaseModel
from typing import Optional
from datetime import date
from typing import Optional

class GWUser(BaseModel):
    id: str
    username: str
    password: str
    role: Optional[str]
    email: Optional[str]
    isactive: Optional[bool]
    registration_date: Optional[date]
    last_login_date: Optional[date]
    loggedIn: Optional[bool]



import inspect
import io
import logging
from functools import wraps
import os
import sys

from pygeoweaver.commands.pgw_history import save_history

logger = logging.getLogger(__name__)

# Global context for tracking
geoweaver_context = {
    'current_workflow': None,
    'process_calls': {}
}


def pygeoweaver_process(func):
    
    @wraps(func)
    def wrapper(*args, **kwargs):
        func_name = func.__name__
        current_file = inspect.getfile(inspect.currentframe())
        current_file = os.path.basename(current_file)
        current_file_without_extension = os.path.splitext(current_file)[0]
        try:
            # Capture the source code of the wrapped function
            func_code = inspect.getsource(func)
        except OSError:
            func_code = f"# Source code not available for {func_name}"

        geoweaver_process_id = f"{current_file_without_extension}.{func_name}"
        logger.info("this process id should be unique: %s", geoweaver_process_id)
        
        # Capture logging output
        logger.info(f"geoweaver captured {func_name}")

        # Capture console output
        captured_stdout = io.StringIO()
        captured_stderr = io.StringIO()
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = captured_stdout
        sys.stderr = captured_stderr
        
        try:
            result = func(*args, **kwargs)
        finally:
            # Restore standard output and error
            sys.stdout = old_stdout
            sys.stderr = old_stderr

            # Log the captured output
            stdout_content = captured_stdout.getvalue()
            stderr_content = captured_stderr.getvalue()
            if stdout_content:
                logger.info(f"Standard Output:\n{stdout_content}")
                print(stdout_content)
            if stderr_content:
                logger.error(f"Standard Error:\n{stderr_content}")
                print(stderr_content)

            logger.info(f"Finished {func_name}")
            
        current_workflow = geoweaver_context['current_workflow']
        if current_workflow:
            if current_workflow not in geoweaver_context['process_calls']:
                geoweaver_context['process_calls'][current_workflow] = []
            geoweaver_context['process_calls'][current_workflow].append({
                'name': func_name,
                'code': func_code,
                'log': f"{stdout_content}\n{stderr_content}"
            })
            
        return result
    
    return wrapper



from pydantic import BaseModel
from datetime import datetime
from typing import Optional

class History(BaseModel):
    history_id: str
    history_input: Optional[str]
    history_output: Optional[str]
    history_begin_time: Optional[datetime]
    history_end_time: Optional[datetime]
    history_notes: Optional[str]
    history_process: Optional[str]
    host_id: Optional[str]
    indicator: Optional[str]



import getpass
from logging import getLogger
import subprocess
from pygeoweaver.utils import (
    check_ipython,
    download_geoweaver_jar,
    get_geoweaver_jar_path,
    get_java_bin_path,
    get_root_dir,
)

logger = getLogger(__name__)

def get_password_twice():
    while True:
        password1 = getpass.getpass('Enter password: ')
        password2 = getpass.getpass('Re-enter password: ')
        
        if password1 == password2:
            return password1
        else:
            print("Passwords don't match. Please try again.")


def reset_password():
    """
    Usage: <main class> resetpassword
    Reset password for localhost
    """
    download_geoweaver_jar()
    
    if check_ipython():
        logger.debug("ipython is here")
        password = get_password_twice()
        subprocess.run(
            [
                get_java_bin_path(),
                "-jar",
                get_geoweaver_jar_path(),
                "resetpassword",
                "-p",
                password
            ]
        )
    else:
        logger.debug("not ipython")
        subprocess.run(
            [
                get_java_bin_path(),
                "-jar",
                get_geoweaver_jar_path(),
                "resetpassword",
            ]
        )



import platform
import os
import subprocess
import shutil
import sys
import tarfile
import zipfile
import urllib.request

import click

from pygeoweaver.utils import detect_rc_file, get_home_dir, get_java_bin_path, safe_exit


def install_jdk():
    system = platform.system()
    architecture = platform.machine()

    if system == "Darwin":
        if architecture == "x86_64":
            install_jdk_macos("11.0.18-10", "jdk_x64_mac_hotspot")
        elif architecture == "arm64":
            install_jdk_macos("11.0.18-10", "jdk_aarch64_mac_hotspot")
        else:
            print("Unsupported architecture.")

    elif system == "Linux":
        if architecture == "x86_64":
            install_jdk_linux("11.0.18-10", "jdk_x64_linux_hotspot")
        elif architecture == "aarch64":
            install_jdk_linux("11.0.18-10", "jdk_aarch64_linux_hotspot")
        else:
            print("Unsupported architecture.")

    elif system == "Windows":
        if architecture == "AMD64" or architecture == "x86_64":
            install_jdk_windows("11.0.18-10", "jdk_x64_windows_hotspot")
        elif architecture == "x86-32":
            install_jdk_windows("11.0.18-10", "jdk_x86-32_windows_hotspot")
        else:
            print("Unsupported architecture.")

    else:
        print("Unsupported platform.")


def install_jdk_macos(jdk_version, jdk_arch):
    # jdk_aarch64_linux_hotspot
    jdk_url = f'https://github.com/adoptium/temurin11-binaries/releases/download/jdk-{jdk_version.replace("-", "%2B")}/OpenJDK11U-{jdk_arch}_{jdk_version.replace("-", "_")}.tar.gz'
    jdk_install_dir = os.path.expanduser("~/jdk")

    # Download JDK archive
    download_file(jdk_url, f"{get_home_dir()}/jdk.tar.gz")

    # Extract JDK archive
    extract_tar_archive(f"{get_home_dir()}/jdk.tar.gz", jdk_install_dir)

    # Set JDK environment variables
    set_jdk_env_vars(f'{jdk_install_dir}/jdk-{jdk_version.replace("-", "+")}')


def install_jdk_linux(jdk_version, jdk_arch):
    jdk_url = f'https://github.com/adoptium/temurin11-binaries/releases/download/jdk-{jdk_version.replace("-", "%2B")}/OpenJDK11U-{jdk_arch}_{jdk_version.replace("-", "_")}.tar.gz'
    jdk_install_dir = os.path.expanduser("~/jdk")

    # Download JDK archive
    download_file(jdk_url, f"{get_home_dir()}/jdk.tar.gz")

    # Extract JDK archive
    extract_tar_archive(f"{get_home_dir()}/jdk.tar.gz", jdk_install_dir)

    # Set JDK environment variables
    set_jdk_env_vars(f'{jdk_install_dir}/jdk-{jdk_version.replace("-", "+")}')


def install_jdk_windows(jdk_version, jdk_arch):
    jdk_url = f'https://github.com/adoptium/temurin11-binaries/releases/download/jdk-{jdk_version.replace("-", "%2B")}/OpenJDK11U-{jdk_arch}_{jdk_version.replace("-", "_")}.zip'
    jdk_install_dir = os.path.expanduser("~/jdk")

    # Download JDK archive
    download_file(jdk_url, f"{get_home_dir()}/jdk.zip")

    # Extract JDK archive
    extract_zip_archive(f"{get_home_dir()}/jdk.zip", jdk_install_dir)

    # Set JDK environment variables
    set_jdk_env_vars(f'{jdk_install_dir}/jdk-{jdk_version.replace("-", "+")}')


def download_file(url, filename):
    if os.path.exists(filename):
        print(f"{filename} already exists.")
        return
    print(f"Downloading {filename}...", url)
    urllib.request.urlretrieve(url, filename)
    print(f"{filename} downloaded.")


def extract_tar_archive(archive_path, destination_dir):
    if not os.path.exists(destination_dir):
        print(f"Extracting {archive_path}...")
        with tarfile.open(archive_path, "r:gz") as tar_ref:
            tar_ref.extractall(destination_dir)
        print(f"{archive_path} extracted to {destination_dir}.")


def extract_zip_archive(archive_path, destination_dir):
    print(f"Extracting {archive_path}...")
    with zipfile.ZipFile(archive_path, "r") as zip_ref:
        zip_ref.extractall(destination_dir)
    print(f"{archive_path} extracted to {destination_dir}.")


def set_jdk_env_vars_for_windows(jdk_install_dir):
    print(f"Setting JDK environment variables for Windows...")
    java_bin = os.path.join(jdk_install_dir, "bin")

    # Append JAVA_HOME to the user's PATH environment variable
    try:
        import winreg  # winreg module is only available on windows and should not be globally imported.

        with winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE,
            "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment",
            0,
            winreg.KEY_ALL_ACCESS,
        ) as regkey:
            current_path = winreg.QueryValueEx(regkey, "PATH")[0]
            print(f"current_path = {current_path}")
            if java_bin not in current_path:
                new_path = f"{current_path};{java_bin}"
                winreg.SetValueEx(regkey, "PATH", 0, winreg.REG_EXPAND_SZ, new_path)
                print(f"Added JDK bin directory to PATH: {java_bin}")
                subprocess.call(["setx", "PATH", ";".join([current_path, java_bin])])
                print("JDK environment variables set.")
    except Exception as e:
        print(f"Error adding JDK bin directory to PATH: {e}")

    # Set JAVA_HOME environment variable
    # try:
    #     with winreg.OpenKey(winreg.HKEY_CURRENT_USER, "Environment", 0, winreg.KEY_ALL_ACCESS) as regkey:
    #         winreg.SetValueEx(regkey, "JAVA_HOME", 0, winreg.REG_EXPAND_SZ, jdk_install_dir)
    #         print("Set JAVA_HOME environment variable.")
    # except Exception as e:
    #     print(f"Error setting JAVA_HOME environment variable: {e}")

    # Update the environment variables of the current process
    # subprocess.call(["setx", "JAVA_HOME", java_home])


def set_jdk_env_vars_for_linux_mac(jdk_install_dir):
    print(f"Setting JDK environment variables...")
    java_line = f'\nexport JAVA_HOME="{jdk_install_dir}"\n'
    rc_file_path = detect_rc_file()

    check_java = False
    with open(rc_file_path, "r") as file:
        for line in file:
            if line.strip() == java_line:
                check_java = True
                break

    if not check_java:
        with open(rc_file_path, "a") as bashrc:
            bashrc.write(f'export JAVA_HOME="{jdk_install_dir}"\n')
            bashrc.write(f'export PATH="$JAVA_HOME/bin:$PATH"\n')
            print("JDK environment variables set.")

    subprocess.run(["bash", "-i", "-c", f"source {rc_file_path} && echo 'Java environment sourced.'"])


def set_jdk_env_vars(jdk_install_dir):
    print(f"Setting environment variables for {platform.system()}")
    if platform.system() == "Windows":
        set_jdk_env_vars_for_windows(jdk_install_dir)
    else:
        set_jdk_env_vars_for_linux_mac(jdk_install_dir)


def install_java():
    system = platform.system()
    if system == "Darwin":
        os.system(
            "/bin/bash -c '/usr/bin/ruby -e \"$(curl -fsSL "
            "https://raw.githubusercontent.com/Homebrew/install/master/install)\"'"
        )
        os.system("brew install openjdk")
    elif system == "Linux":
        # need to check if the package manager type is apt or yum
        # arch / debian
        package_manager = None
        if os.path.exists("/usr/bin/apt"):
            package_manager = "apt"
        elif os.path.exists("/usr/bin/yum"):
            package_manager = "yum"

        if package_manager:
            os.system(f"sudo {package_manager} update")
            os.system(f"sudo {package_manager} install -y default-jre default-jdk")
        else:
            print("Package manager not found. Unable to install Java.")
            safe_exit(1)
    elif system == "Windows":
        # note: this requires admin access to the pc, else it will fail saying
        # Access to the path 'C:\ProgramData\chocolatey\lib-bad' is denied.
        os.system(
            'powershell -Command "Set-ExecutionPolicy Bypass -Scope Process -Force; ['
            "System.Net.ServicePointManager]::SecurityProtocol = 3072; iex ((New-Object "
            "System.Net.WebClient).DownloadString('https://chocolatey.org/install.ps1'))\""
        )
        os.system("choco install -y openjdk")
    else:
        print("Unsupported operating system.")
        safe_exit(1)


def is_java_installed():
    try:
        # Check if Java is installed by running "java -version" command
        subprocess.run(
            [get_java_bin_path(), "-version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return True
    except:
        return False


def check_java():
    # Check if Java is installed
    if not is_java_installed():
        click.echo(click.style("Java is not installed. Installing...", fg="yellow"))
        try:
            install_jdk()
            if is_java_installed():
                click.echo(click.style("Java installation complete.", fg="green"))
            else:
                click.echo(click.style("Java is still not installed correctly. Please follow the instructions below.", fg="red"))
                click.echo(click.style("Step 1: Visit https://adoptium.net/ to download and install OpenJDK (Eclipse Adoptium).", fg="blue"))
                click.echo(click.style("Step 2: Choose the appropriate JDK version for your operating system.", fg="blue"))
                click.echo(click.style("Step 3: Follow the installation instructions provided on the website.", fg="blue"))
                click.echo(click.style("Step 4: After installation, make sure JAVA_HOME and PATH are set correctly.", fg="blue"))
                click.echo(click.style("If you encounter any problems, please contact support at geoweaver.app@gmail.com or post in GitHub issues: https://github.com/ESIPFed/pygeoweaver/issues.", fg="red"))
        except Exception as e:
            click.echo(click.style(f"Error: {e}", fg="red"))
            click.echo(click.style("Please contact support at support@example.com for further assistance.", fg="red"))



"""
All functions of subcommand interfaces
"""
from pygeoweaver.commands.pgw_detail import *
from pygeoweaver.commands.pgw_list import *
from pygeoweaver.commands.pgw_export import *
from pygeoweaver.commands.pgw_history import *
from pygeoweaver.commands.pgw_import import *
from pygeoweaver.commands.pgw_list import *
from pygeoweaver.commands.pgw_run import *
from pygeoweaver.server import *
from pygeoweaver.commands.pgw_resetpassword import *
from pygeoweaver.commands.pgw_help import *
from pygeoweaver.commands.pgw_create import *
from pygeoweaver.commands.pgw_sync import *
from pygeoweaver.commands.pgw_find import *
from pygeoweaver.commands.pgw_cleanh2db import *
from pygeoweaver.commands.pgw_upgrade import *



import json
import logging

import requests
import subprocess
from pygeoweaver.constants import *
from pygeoweaver.utils import (
    download_geoweaver_jar,
    get_geoweaver_jar_path,
    get_java_bin_path,
    get_root_dir,
    check_ipython,
    get_spinner,
)
import pandas as pd
from halo import Halo


logger = logging.getLogger(__name__)


def list_hosts():
    """
    List all hosts in Geoweaver.

    Downloads the Geoweaver JAR, then runs the 'list' command with the '--host' option.

    Note: Requires Geoweaver to be initialized and the JAR file to be available.
    """
    with get_spinner(text=f'Find all registered hosts...', spinner='dots'):
        download_geoweaver_jar()
        process = subprocess.run(
            [get_java_bin_path(), "-jar", get_geoweaver_jar_path(), "list", "--host"],
            cwd=f"{get_root_dir()}/",
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )

    print(process.stdout)
    if process.stderr:
        print("=== Error ===")
        print(process.stderr)
        logger.error(process.stderr)


def list_processes():
    """
    List all processes in Geoweaver.

    Downloads the Geoweaver JAR, ensures executable permissions, and then runs the 'list' command with the '--process' option.

    Note: Requires Geoweaver to be initialized and the JAR file to be available.
    """
    with get_spinner(text=f'Find all registered processes...', spinner='dots'):
        download_geoweaver_jar()
        subprocess.run(["chmod", "+x", get_geoweaver_jar_path()], cwd=f"{get_root_dir()}/")
        process = subprocess.run(
            [get_java_bin_path(), "-jar", get_geoweaver_jar_path(), "list", "--process"],
            cwd=f"{get_root_dir()}/",
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )

    print(process.stdout)
    if process.stderr:
        print("=== Error ===")
        print(process.stderr)
        logger.error(process.stderr)


def list_processes_in_workflow(workflow_id):
    """
    List processes in a specific workflow.

    Downloads the Geoweaver JAR and queries the Geoweaver server for details of a workflow.
    Extracts information about nodes in the workflow and returns a list of dictionaries containing 'title' and 'id'.

    :param workflow_id: The ID of the workflow.
    :type workflow_id: str
    :return: List of dictionaries containing 'title' and 'id' of processes in the workflow.
    :rtype: list
    """
    with get_spinner(text=f'Find all processes in workflow {workflow_id}...', spinner='dots'):
        download_geoweaver_jar()
        payload = {"id": workflow_id, "type": "workflow"}
        r = requests.post(
            f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/detail", data=payload
        )
        nodes = json.loads(r.json()["nodes"])
        result = [
            {"title": item["title"], "id": item["id"].split(".")[0]} for item in nodes
        ]

    if check_ipython():
        return pd.DataFrame(result)
    return result


def list_workflows():
    """
    List all workflows in Geoweaver.

    Downloads the Geoweaver JAR, then runs the 'list' command with the '--workflow' option.

    Note: Requires Geoweaver to be initialized and the JAR file to be available.
    """
    with get_spinner(text=f'Find all registered workflows...', spinner='dots'):
        download_geoweaver_jar()
        process = subprocess.run(
            [get_java_bin_path(), "-jar", get_geoweaver_jar_path(), "list", "--workflow"],
            cwd=f"{get_root_dir()}/",
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )
    
    print(process.stdout)
    if process.stderr:
        print("=== Error ===")
        print(process.stderr)
        logger.error(process.stderr)




import os
import re
import shutil
import click
from pathlib import Path

def cleanup_workspace():
    """Clean up temporary directories in the gw-workspace folder."""
    try:
        # Get the user's home directory in a cross-platform way
        home_dir = str(Path.home())
        workspace_dir = os.path.join(home_dir, "gw-workspace")
        
        if not os.path.exists(workspace_dir):
            click.echo(f"Workspace directory {workspace_dir} does not exist.")
            return
        
        # Compile regex patterns for 11 and 12 character directory names
        pattern_11 = re.compile(r'^[a-zA-Z0-9]{11}$')
        pattern_12 = re.compile(r'^[a-zA-Z0-9]{12}$')
        
        # Counter for removed directories
        removed_count = 0
        
        # Iterate through immediate subdirectories
        for item in os.listdir(workspace_dir):
            item_path = os.path.join(workspace_dir, item)
            
            # Check if it's a directory and matches either pattern
            if os.path.isdir(item_path) and (pattern_11.match(item) or pattern_12.match(item)):
                try:
                    shutil.rmtree(item_path)
                    removed_count += 1
                except Exception as e:
                    click.echo(f"Error removing directory {item}: {str(e)}")
        
        if removed_count > 0:
            click.echo(f"Successfully removed {removed_count} temporary directories.")
        else:
            click.echo("No temporary directories found to clean up.")
            
    except Exception as e:
        click.echo(f"Error during cleanup: {str(e)}")


import re
from unittest.mock import patch
from pygeoweaver.commands.pgw_detail import (
    detail_process,
    detail_workflow,
    detail_host,
)


def clean_output(output):
    """
    Clean escape sequences and spinner content from the output.
    """
    return re.sub(r"\x1b\[.*?m|\r|\⠋.*?\⠙.*?", "", output).strip()


# Correct mock path: Mock check_geoweaver_status where it is used (pgw_detail)
@patch("pygeoweaver.commands.pgw_detail.check_geoweaver_status", return_value=True)
@patch("pygeoweaver.commands.pgw_detail.ensure_server_running", return_value=None)
def test_detail_process(mock_ensure_server, mock_check_status, capfd):
    """
    Test the detail_process function with a non-existing process ID.
    """
    detail_process("not_existing_id")
    output, err = capfd.readouterr()
    clean_out = clean_output(output)
    assert "Error" in clean_out or "not_existing_id" in clean_out


@patch("pygeoweaver.commands.pgw_detail.check_geoweaver_status", return_value=True)
@patch("pygeoweaver.commands.pgw_detail.ensure_server_running", return_value=None)
def test_detail_workflow(mock_ensure_server, mock_check_status, capfd):
    """
    Test the detail_workflow function with a non-existing workflow ID.
    """
    detail_workflow("not_existing_id")
    output, err = capfd.readouterr()
    clean_out = clean_output(output)
    assert "Error" in clean_out or "not_existing_id" in clean_out


@patch("pygeoweaver.commands.pgw_detail.check_geoweaver_status", return_value=True)
@patch("pygeoweaver.commands.pgw_detail.ensure_server_running", return_value=None)
def test_detail_host(mock_ensure_server, mock_check_status, capfd):
    """
    Test the detail_host function with a non-existing host ID.
    """
    detail_host("not_existing_id")
    output, err = capfd.readouterr()
    clean_out = clean_output(output)
    assert "Error" in clean_out or "not_existing_id" in clean_out



import http
import json
import logging
import subprocess
import traceback
from urllib.parse import urlencode, urlparse

import click
import pandas as pd
import requests
from tabulate import tabulate

from pygeoweaver.constants import *
from pygeoweaver.server import check_geoweaver_status, start
from pygeoweaver.utils import (
    download_geoweaver_jar,
    get_geoweaver_jar_path,
    get_java_bin_path,
    get_root_dir,
    get_spinner,
    is_interactive,
)

logger = logging.getLogger(__name__)


def display_response_table(response_json):
    """
    Convert the response JSON into a table format and display it.
    
    Args:
        response_json (str or dict): The JSON response to display.
    """
    if isinstance(response_json, str):
        data = json.loads(response_json)
    else:
        data = response_json

    # Convert data into a DataFrame for better visualization
    df = pd.DataFrame([data])

    # Check if running in Jupyter
    if is_interactive():
        from IPython.display import display
        display(df)
    else:
        # Use tabulate for terminal display
        table = tabulate(df, headers='keys', tablefmt='psql')
        print(table)
        


def show_history(history_id):
    """
    History uses the same method to check
    """
    if not history_id:
        raise RuntimeError("history id is missing")
    
    try:
        if not check_geoweaver_status():
            start()
        
        log_url = f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/log"
        json_data = f"type=process&id={history_id}"
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'
        }
        response = requests.post(log_url, headers=headers, data=json_data)

        if response.status_code == 200:
            try:
                display_response_table(response.json())
            except requests.exceptions.JSONDecodeError:
                raise ValueError('Response is not in JSON format:', response.text)
        else:
            raise ValueError(f'POST request failed with status code {response.text}')
    except Exception as e:
        with get_spinner(text=f'Get workflow history...', spinner='dots'):
            logger.error(e)
            traceback_str = traceback.format_exc()
            print(f"Error occurred: {str(e)}")
            print(f"Traceback:\n{traceback_str}")
            
            download_geoweaver_jar()
            process = subprocess.run(
                [get_java_bin_path(), "-jar", get_geoweaver_jar_path(), "history", history_id],
                cwd=f"{get_root_dir()}/",
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
            )
        
        print(process.stdout)
        if process.stderr:
            print("=== Error ===")
            print(process.stderr)
            logger.error(process.stderr)


def get_process_history(process_id):
    """
        Get list of history for a process using process id
    :param process_id: str    :type process_id: str
    """
    if not process_id:
        raise Exception("please pass `process_id` as a parameter to the function.")
    try:
        print("why it is still here")
        if check_geoweaver_status():
            start(exit_on_finish=False)
            print("why it is still here")
        
        json_data = f"type=process&id={process_id}"
        print(json_data)
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'
        }
        r = requests.post(
            f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/logs",
            data=json_data,
            headers=headers,
        )
        print("r.text = ", r.text)
        if r.status_code == 200:
            try:
                display_response_table(r.json())
            except requests.exceptions.JSONDecodeError:
                raise ValueError('Response is not in JSON format:', r.text)
        else:
            raise ValueError(f'POST request failed with status code {r.text}')
    except Exception as e:
        with get_spinner(text=f'Get workflow history...', spinner='dots'):
            process = subprocess.run(
                f"{get_java_bin_path()} -jar {get_geoweaver_jar_path()} process-history {process_id}",
                cwd=f"{get_root_dir()}/",
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
            )
        
        print(process.stdout)
        if process.stderr:
            print("=== Error ===")
            print(process.stderr)
            logger.error(process.stderr)


def get_workflow_history(workflow_id):
    """
        Get list of history for a workflow using workflow id
    :param workflow_id: str
    """
    if not workflow_id:
        raise Exception("please pass `workflow_id` as a parameter to the function.")
    try:
        if check_geoweaver_status():
            start()
        r = requests.post(
            f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/logs",
            data={"type": "workflow", "id": workflow_id},
        ).json()
        df = pd.DataFrame(r)
        df["history_begin_time"] = pd.to_datetime(df["history_begin_time"], unit="ms")
        df["history_end_time"] = pd.to_datetime(df["history_end_time"], unit="ms")
        if is_interactive():
            return df
        else:
            print(df)
    except Exception as e:
        with get_spinner(text=f'Get workflow history via slow CLI...', spinner='dots'):
            process = subprocess.run(
                f"{get_java_bin_path()} -jar {get_geoweaver_jar_path()} workflow-history {workflow_id}",
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, 
                cwd=f"{get_root_dir()}/",
            )
            
        print(process.stdout)
        if process.stderr:
            print("=== Error ===")
            print(process.stderr)
            logger.error(process.stderr)


def save_history(code: str = None, status: str = None, log_output: str = None, ):
    
    pass



"""Utility functions for handling configuration files in Geoweaver."""

import os
import logging
from configparser import ConfigParser

logger = logging.getLogger(__name__)

def read_properties_file(file_path):
    """
    Read a Java-style properties file and return a dictionary of key-value pairs.
    
    Args:
        file_path (str): Path to the properties file
        
    Returns:
        dict: Dictionary containing the properties
    """
    properties = {}
    
    if not os.path.exists(file_path):
        logger.debug(f"Properties file not found: {file_path}")
        return properties
    
    try:
        with open(file_path, 'r') as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue
                    
                if '=' in line:
                    key, value = line.split('=', 1)
                    properties[key.strip()] = value.strip()
    except Exception as e:
        logger.error(f"Error reading properties file {file_path}: {e}")
    
    return properties

def get_database_url_from_properties():
    """
    Get the database URL from the application.properties file if it exists.
    
    Returns:
        str or None: The database URL if found, None otherwise
    """
    home_dir = os.path.expanduser("~")
    properties_path = os.path.join(home_dir, "geoweaver", "application.properties")
    
    if not os.path.exists(properties_path):
        logger.info(f"Application properties file not found at {properties_path}")
        return None
    
    logger.info(f"Reading database configuration from {properties_path}")
    properties = read_properties_file(properties_path)
    
    # Look for database URL configuration
    # Common property names for database URL in Spring applications
    possible_keys = [
        "spring.datasource.url",
        "database.url",
        "db.url",
        "datasource.url",
        "jdbc.url"
    ]
    
    for key in possible_keys:
        if key in properties:
            url = properties[key]
            logger.info(f"Found database URL in properties file: {url}")
            return url
    
    logger.info("No database URL configuration found in properties file")
    return None


"""Module for cleaning and reducing the size of the H2 database used by Geoweaver."""

import logging
import os
import shutil
import subprocess
import re
import urllib.request
import platform
import tempfile
from pathlib import Path

from pygeoweaver.server import stop, start, check_geoweaver_status
from pygeoweaver.utils import get_spinner, safe_exit, get_home_dir
from pygeoweaver.jdk_utils import check_java, download_file
from pygeoweaver.config_utils import get_database_url_from_properties

logger = logging.getLogger(__name__)

def clean_h2db(h2_jar_path=None, temp_dir=None, db_path=None, username="geoweaver", password=None):
    """
    Clean and reduce the size of the H2 database used by Geoweaver.
    
    This function follows these steps:
    1. Stop Geoweaver if it's running
    2. Create a temporary directory if one is not provided
    3. Copy database files to the temporary directory
    4. Export data from the database to a SQL file
    5. Remove the original database files
    6. Import the SQL file into a new database
    7. Start Geoweaver
    
    Args:
        h2_jar_path (str, optional): Path to the H2 database JAR file. If not provided, will use h2-2.2.224.jar in the current directory.
        temp_dir (str, optional): Path to a temporary directory for the recovery process. If not provided, will create one.
        db_path (str, optional): Path to the H2 database files. If not provided, will use ~/h2_hopper_amd_1/gw.
        username (str, optional): Username for the H2 database. Defaults to "geoweaver".
        password (str, optional): Password for the H2 database. If not provided, will prompt the user.
    
    Returns:
        bool: True if the operation was successful, False otherwise.
    """
    # Check if Java is installed
    check_java()
    
    # Step 1: Stop Geoweaver if it's running
    with get_spinner(text="Checking if Geoweaver is running...", spinner="dots"):
        if check_geoweaver_status():
            logger.info("Stopping Geoweaver...")
            stop(exit_on_finish=False)
        else:
            logger.info("Geoweaver is not running.")
    
    # Set default paths if not provided
    home_dir = os.path.expanduser("~")
    if not db_path:
        # Check if there's a custom database URL in application.properties
        custom_db_url = get_database_url_from_properties()
        if custom_db_url and "jdbc:h2:" in custom_db_url:
            # Extract the file path from the JDBC URL
            # Format is typically: jdbc:h2:file:/path/to/database
            # or jdbc:h2:/path/to/database
            match = re.search(r'jdbc:h2:(?:file:)?(.+?)(?:;|$)', custom_db_url)
            if match:
                db_path = match.group(1)
                logger.info(f"Using database path from application.properties: {db_path}")
            else:
                logger.warning(f"Could not parse database path from URL: {custom_db_url}")
                # Use default path with 'h2' instead of 'h2_hopper_amd_1'
                db_path = os.path.join(home_dir, "h2", "gw")
        else:
            # Use default path with 'h2' instead of 'h2_hopper_amd_1'
            db_path = os.path.join(home_dir, "h2", "gw")
        
        logger.info(f"Using database path: {db_path}")
    
    # Set default H2 JAR path and version
    h2_version = "2.2.224"
    if not h2_jar_path:
        # First check in current directory
        h2_jar_path = f"h2-{h2_version}.jar"
        
        # If not in current directory, check in home directory
        if not os.path.exists(h2_jar_path):
            home_h2_jar_path = os.path.join(get_home_dir(), f"h2-{h2_version}.jar")
            
            # If not in home directory, download it
            if not os.path.exists(home_h2_jar_path):
                with get_spinner(text=f"Downloading H2 database JAR file (version {h2_version})...", spinner="dots"):
                    try:
                        # URL for the H2 database JAR file
                        h2_download_url = f"https://repo1.maven.org/maven2/com/h2database/h2/{h2_version}/h2-{h2_version}.jar"
                        
                        # Download the file
                        download_file(h2_download_url, home_h2_jar_path)
                        logger.info(f"H2 JAR file downloaded to {home_h2_jar_path}")
                    except Exception as e:
                        logger.error(f"Failed to download H2 JAR file: {str(e)}")
                        return False
            
            # Use the JAR file in the home directory
            h2_jar_path = home_h2_jar_path
            
        logger.info(f"Using H2 JAR file at {h2_jar_path}")
        
        # Final check to ensure the JAR file exists
        if not os.path.exists(h2_jar_path):
            logger.error(f"H2 JAR file not found at {h2_jar_path} and could not be downloaded")
            return False
    
    # Create a temporary directory if not provided
    if not temp_dir:
        # Try to use system temp directory first as it usually has more space
        system_temp = tempfile.gettempdir()
        username = os.environ.get('USER', os.environ.get('USERNAME', 'user'))
        
        # Check if we're on a cluster system that might have a scratch directory
        potential_dirs = [
            # Standard system temp directory
            system_temp,
            # User-specific scratch directory on clusters
            f"/scratch/{username}",
            # Alternative scratch locations
            f"/tmp/{username}",
            # Fallback to home directory
            os.path.join(home_dir, "geoweaver", "h2")
        ]
        
        # Find the first directory that exists and has sufficient space
        for potential_dir in potential_dirs:
            if os.path.exists(os.path.dirname(potential_dir)):
                try:
                    # Check available disk space (in bytes)
                    if platform.system() == "Windows":
                        free_bytes = shutil.disk_usage(os.path.dirname(potential_dir)).free
                    else:
                        # Unix systems
                        stat = os.statvfs(os.path.dirname(potential_dir))
                        free_bytes = stat.f_frsize * stat.f_bavail
                    
                    # Require at least 1GB of free space (adjust as needed)
                    required_space = 1 * 1024 * 1024 * 1024  # 1GB in bytes
                    
                    if free_bytes > required_space:
                        temp_dir = os.path.join(potential_dir, "geoweaver_h2_temp")
                        logger.info(f"Using temporary directory with {free_bytes / (1024**3):.2f}GB free space: {temp_dir}")
                        break
                    else:
                        logger.warning(f"Insufficient disk space in {potential_dir}: {free_bytes / (1024**3):.2f}GB free")
                except Exception as e:
                    logger.warning(f"Error checking disk space in {potential_dir}: {str(e)}")
                    continue
        
        # If no suitable directory was found, use system temp as last resort
        if not temp_dir:
            temp_dir = os.path.join(system_temp, f"geoweaver_h2_temp_{username}")
            logger.warning(f"No directory with sufficient space found, using system temp as last resort: {temp_dir}")
    
    # Ensure the temporary directory exists
    os.makedirs(temp_dir, exist_ok=True)
    
    logger.info(f"Using temporary directory: {temp_dir}")
    
    # Step 2: Copy database files to the temporary directory
    with get_spinner(text="Copying database files to temporary directory...", spinner="dots"):
        db_dir = os.path.dirname(db_path)
        
        # Check if the directory exists before trying to list its contents
        if not os.path.exists(db_dir):
            logger.warning(f"Database directory does not exist: {db_dir}")
            logger.info(f"Creating directory: {db_dir}")
            os.makedirs(db_dir, exist_ok=True)
            # Since the directory didn't exist, there are no files to copy
            logger.info(f"No database files to copy from {db_dir}")
        else:
            # Directory exists, copy any matching files
            files_copied = False
            for file in os.listdir(db_dir):
                if file.startswith(os.path.basename(db_path)):
                    src_file = os.path.join(db_dir, file)
                    dst_file = os.path.join(temp_dir, file)
                    shutil.copy2(src_file, dst_file)
                    logger.info(f"Copied {src_file} to {dst_file}")
                    files_copied = True
            
            if not files_copied:
                logger.info(f"No database files found in {db_dir} matching pattern {os.path.basename(db_path)}*")
    
    # Step 3: Export data from the database to a SQL file
    sql_file = os.path.join(temp_dir, "gw_backup.sql")
    
    # Prompt for password if not provided - moved outside of spinner context
    if not password:
        import getpass
        password = getpass.getpass("Enter database password: ")
        
    with get_spinner(text="Exporting database to SQL file...", spinner="dots"):
        # Build the command to export the database
        export_cmd = [
            "java", "-cp", h2_jar_path, 
            "org.h2.tools.Script", 
            "-url", f"jdbc:h2:{os.path.join(temp_dir, os.path.basename(db_path))}", 
            "-user", username, 
            "-script", sql_file, 
            "-password", password
        ]
        
        try:
            subprocess.run(export_cmd, check=True, capture_output=True, text=True)
            logger.info(f"Database exported to {sql_file}")
        except subprocess.CalledProcessError as e:
            logger.error(f"Failed to export database: {e.stderr}")
            return False
    
    # Step 4: Remove the original database files
    with get_spinner(text="Removing original database files...", spinner="dots"):
        # Check if the directory exists before trying to remove files
        if not os.path.exists(db_dir):
            logger.warning(f"Database directory does not exist: {db_dir}")
            # Create the directory for the import step
            os.makedirs(db_dir, exist_ok=True)
        else:
            # Directory exists, remove any matching files
            files_removed = False
            for file in os.listdir(db_dir):
                if file.startswith(os.path.basename(db_path)):
                    os.remove(os.path.join(db_dir, file))
                    logger.info(f"Removed {os.path.join(db_dir, file)}")
                    files_removed = True
            
            if not files_removed:
                logger.info(f"No database files found to remove in {db_dir}")
    
    # Step 5: Import the SQL file into a new database
    with get_spinner(text="Importing SQL file into new database...", spinner="dots"):
        # Ensure the target directory exists
        db_dir = os.path.dirname(db_path)
        if not os.path.exists(db_dir):
            logger.info(f"Creating database directory: {db_dir}")
            os.makedirs(db_dir, exist_ok=True)
            
        # Build the command to import the database
        import_cmd = [
            "java", "-cp", h2_jar_path, 
            "org.h2.tools.RunScript", 
            "-url", f"jdbc:h2:{db_path}", 
            "-user", username, 
            "-script", sql_file, 
            "-password", password
        ]
        
        try:
            subprocess.run(import_cmd, check=True, capture_output=True, text=True)
            logger.info(f"Database imported from {sql_file}")
        except subprocess.CalledProcessError as e:
            logger.error(f"Failed to import database: {e.stderr}")
            return False
    
    # Step 6: Start Geoweaver
    with get_spinner(text="Starting Geoweaver...", spinner="dots"):
        start(exit_on_finish=False)
    
    print("\nH2 database cleanup completed successfully!")
    print(f"Temporary files are stored in {temp_dir}")
    print("You can delete these files if the database is working correctly.")
    
    return True



import requests

from pygeoweaver.api_call.pgw_base_api_caller import BaseAPI

class GeoweaverProcessAPI(BaseAPI):
    
    def edit_process(self, process_data):
        return self._call_api('/edit/process', method='POST', data=process_data)

    def add_process(self, process_data):
        return self._call_api('/add/process', method='POST', data=process_data)

    def get_process(self, process_id):
        return self._call_api(f'/get/process/{process_id}', method='GET')

    def delete_process(self, process_id):
        return self._call_api(f'/delete/process/{process_id}', method='DELETE')




import unittest
import subprocess
import sys

class TestPythonInstallation(unittest.TestCase):

  def test_installation(self):
    # Attempt to install your package
    result = subprocess.run([sys.executable, '-m', 'pip', 'install', 'pygeoweaver'],
                            capture_output=True, text=True)

    self.assertEqual(result.returncode, 0, "Installation failed")

    # Optionally, try importing your package
    try:
      import pygeoweaver
    except ImportError:
      self.fail("Failed to import pygeoweaver after installation")

if __name__ == '__main__':
  unittest.main()



import os
import sys
import shutil
import logging
import subprocess
import requests
import platform

from IPython import get_ipython
from halo import Halo
from pygeoweaver.constants import GEOWEAVER_URL
from pygeoweaver.pgw_log_config import get_logger
from pygeoweaver.pgw_spinner import Spinner

logger = get_logger(__name__)

def is_interactive():
    """
    Check if the code is running in an interactive environment like Jupyter Notebook.
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


def get_spinner(text: str, spinner: str = "dots"):
    if is_interactive():
        return Spinner(text=text, style=spinner)
    else:
        return Halo(text=text, spinner=spinner)


def safe_exit(code=0):
    """
    Safely exit the script or notebook session.

    Parameters:
    - code (int): Exit status code (default: 0 for success).
    """
    if 'ipykernel' in sys.modules:
        # Running in Jupyter notebook or IPython
        # don't exit at all in Jupyter
        pass
    else:
        # Running in a terminal or other environment
        sys.exit(code)


def get_home_dir():
    """
    Get the user's home directory.
    """
    if platform.system() == "Windows":
        return os.path.expandvars("%USERPROFILE%")
    else:
        return os.path.expanduser("~")


def get_root_dir():
    """
    Get the root directory of the module.
    """
    head, tail = os.path.split(__file__)
    return head


def detect_rc_file():
    """
    Detect the user's shell and return the appropriate shell configuration (rc) file path.
    Create the rc file if it doesn't exist.
    
    Returns:
        str: Path to the shell configuration file (e.g., .bashrc, .zshrc, config.fish).
    """
    # Detect user's shell
    user_shell = os.environ.get('SHELL', '/bin/bash')
    
    # Determine appropriate shell configuration file based on the detected shell
    if 'bash' in user_shell:
        rc_file = os.path.expanduser("~/.bashrc")
    elif 'zsh' in user_shell:
        rc_file = os.path.expanduser("~/.zshrc")
    elif 'fish' in user_shell:
        rc_file = os.path.expanduser("~/.config/fish/config.fish")
    else:
        # Default to bashrc if unknown shell
        rc_file = os.path.expanduser("~/.bashrc")
    
    # Check if the shell configuration file exists, create if it doesn't
    if not os.path.exists(rc_file):
        print(f"{rc_file} does not exist. Creating it...")
        # Ensure the directory exists (for fish, the config directory may not exist)
        os.makedirs(os.path.dirname(rc_file), exist_ok=True)
        open(rc_file, 'a').close()
    
    return rc_file


def get_java_bin_from_which():
    """
    Get the path of the Java binary using the 'which' command.
    """
    system = platform.system()

    if system == "Darwin" or system == "Linux":
        try:
            # Source ~/.bashrc (Assuming it's a non-login shell)
            bashrc_path = detect_rc_file()
            subprocess.run(["bash", "-c", f"source {bashrc_path}"])

            # Check the location of Java executable
            result = subprocess.run(["which", "java"], capture_output=True, text=True)
            java_bin_path = result.stdout.strip()
        except subprocess.CalledProcessError as e:
            print(f"Command execution failed: {e.output}")
            return None
    elif system == "Windows":
        # Check the location of Java executable
        result = subprocess.run(["where", "java"], capture_output=True, text=True)
        java_bin_path = result.stdout.strip()
        
    else:
        print("Unsupported platform.")

    return java_bin_path


def check_java_in_default_env(java_exe="java"):
    try:
        # Attempt to run 'java -version' in the default environment
        subprocess.run([java_exe, "-version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        return True
    except FileNotFoundError:
        # If 'java' is not found in the default environment
        return False
    except subprocess.CalledProcessError:
        # If 'java' is found but there is an issue with execution
        return False

def get_java_bin_path(java_exe="java"):
    java_bin_path = None
    home_dir = get_home_dir()

    # First check if Java is available in the default environment
    if check_java_in_default_env(java_exe):
        # If Java is found in the default environment, return its path
        logger.info(f"Java is available in the default environment.")
        return java_exe

    # Check if JAVA_HOME is set
    java_home = os.environ.get("JAVA_HOME")
    if java_home:
        java_cmd = os.path.join(java_home, "bin", java_exe)
        if os.path.exists(java_cmd):
            java_bin_path = java_cmd
            print(f"Found Java in JAVA_HOME: {java_bin_path}")
            return java_bin_path

    # Look for common installation paths
    common_paths = [
        os.path.join(home_dir, "jdk"),
        os.path.join(home_dir, "java"),
        "/usr/lib/jvm",  # Common location on Linux
        "/usr/bin/java",  # Alternate location on Linux
        "/usr/local/bin/java",
        "C:\\Program Files\\Java",  # Common location on Windows
        "C:\\Program Files (x86)\\Java",  # Alternate location on Windows
    ]

    for base_path in common_paths:
        if os.path.exists(base_path):
            for root, dirs, _ in os.walk(base_path):
                if "bin" in dirs:
                    java_cmd = os.path.join(root, "bin", java_exe)
                    if os.path.exists(java_cmd):
                        java_bin_path = java_cmd
                        print(f"Found Java in {base_path}: {java_bin_path}")
                        return java_bin_path

    # If Java is still not found, we can download and install it to the home directory
    print("Java not found, proceeding to download and install it to the home directory.")
    # Add code to download and install Java here (for example using wget, curl, or a package manager)
    # For now, let's assume the installation is handled and return the path after installation.
    java_bin_path = os.path.join(home_dir, "java", "bin", java_exe)
    return java_bin_path

def get_module_absolute_path():
    """
    Get the absolute path of the module.
    """
    module_path = os.path.abspath(__file__)
    return os.path.dirname(module_path)


def get_geoweaver_jar_path():
    """
    Get the path of the Geoweaver JAR file.
    """
    return f"{get_home_dir()}/geoweaver.jar"


def check_geoweaver_jar():
    """
    Check if the Geoweaver JAR file exists.
    """
    return os.path.isfile(get_geoweaver_jar_path())


def download_geoweaver_jar(overwrite=False):
    """
    Download the latest version of Geoweaver JAR file.
    """
    with get_spinner(text='Checking Geoweaver JAR file...', spinner='dots'):
        if check_geoweaver_jar():
            if overwrite:
                os.remove(get_geoweaver_jar_path())
            else:
                system = platform.system()
                if not system == "Windows":  # Windows files are exec by default
                    subprocess.run(
                        ["chmod", "+x", get_geoweaver_jar_path()], cwd=f"{get_root_dir()}/"
                    )
                return

    with get_spinner(text='Downloading latest version of Geoweaver...', spinner='dots'):
        r = requests.get(GEOWEAVER_URL)
        with open(get_geoweaver_jar_path(), "wb") as f:
            f.write(r.content)

        if check_geoweaver_jar():
            print("Geoweaver.jar is downloaded")
        else:
            raise RuntimeError("Fail to download geoweaver.jar")


def check_os():
    """
    Check the operating system and return corresponding code.
    1: Linux, 2: MacOS, 3: Windows
    """
    if platform.system() == "Linux" or platform == "Linux2":
        return 1
    elif platform.system() == "Darwin":
        return 2
    elif platform.system() == "Windows":
        return 3


def check_ipython():
    """
    Check if the code is running in an IPython environment.
    """
    try:
        return get_ipython().__class__.__name__ == "ZMQInteractiveShell"
    except:
        return False


def copy_files(source_folder, destination_folder):
    """
    Copy files from the source folder to the destination folder.
    """
    for root, dirs, files in os.walk(source_folder):
        for file in files:
            source_file = os.path.join(root, file)
            destination_file = os.path.join(
                destination_folder, os.path.relpath(source_file, source_folder)
            )
            os.makedirs(os.path.dirname(destination_file), exist_ok=True)
            shutil.copy2(source_file, destination_file)

def get_geoweaver_port():
    return os.getenv("GEOWEAVER_PORT", "8070")

def get_log_file_path():
    """
    Determine the best location to store the geoweaver log file
    based on the operating system.
    
    Returns:
        str: The full path to the geoweaver log file.
    """
    # Get the user's home directory
    home_dir = os.path.expanduser("~")
    
    # Determine the log directory based on the platform
    if os.name == 'nt':  # Windows
        log_dir = os.path.join(os.getenv('APPDATA'), 'Geoweaver', 'logs')
    else:  # macOS/Linux
        log_dir = os.path.join(home_dir, '.local', 'share', 'geoweaver', 'logs')
    
    # Create the directory if it doesn't exist
    os.makedirs(log_dir, exist_ok=True)
    
    # Define the full path to the log file
    log_file = os.path.join(log_dir, "geoweaver.log")
    
    # Ensure the log file exists, create it if it doesn't
    if not os.path.exists(log_file):
        open(log_file, "a").close()  # Create an empty log file if it doesn't exist
    
    return log_file




import os
import socket
import subprocess
import sys
import webbrowser
import time
import psutil
import requests
from halo import Halo

from pygeoweaver.constants import GEOWEAVER_DEFAULT_ENDPOINT_URL
from pygeoweaver.jdk_utils import check_java
from pygeoweaver.pgw_log_config import get_logger
from pygeoweaver.utils import (
    check_ipython,
    check_os,
    download_geoweaver_jar,
    get_log_file_path,
    get_module_absolute_path,
    get_root_dir,
    get_spinner,
    safe_exit,
)

"""
This module provides function to start and stop Geoweaver server.
If it detects the current environment is Jupyter notebook, it will 
open Geoweaver GUI in the output cell (if gui is not disabld.)

"""

logger = get_logger(__name__)

# Get the user's home directory
home_dir = os.path.expanduser("~")


def check_geoweaver_status() -> bool:
    """
    Check if geoweaver is running
    """
    try:
        # Run 'ps' command to list all processes
        ps_output = subprocess.check_output(['ps', 'aux']).decode('utf-8').splitlines()
        
        # Check each line of ps output for 'geoweaver.jar'
        geoweaver_running = False
        for line in ps_output:
            if 'geoweaver.jar' in line:
                geoweaver_running = True
                break
        
        if geoweaver_running:
            logger.info("Geoweaver is running.")
            return True
        else:
            logger.info("Geoweaver is not running.")
            return False
    
    except subprocess.CalledProcessError as e:
        err_msg = f"Error checking Geoweaver status: {e}"
        logger.error(err_msg)
        raise ValueError(err_msg)


def start_on_windows(force_restart=False, force_download=False, exit_on_finish=True):
    
    with get_spinner(text=f'Stop running Geoweaver if any...', spinner='dots'):
        subprocess.run(["taskkill", "/f", "/im", "geoweaver.exe"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    with get_spinner(text=f'Check if Java is installed...', spinner='dots'):
        java_cmd = "java"
        try:
            subprocess.run(["where", "java"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)
        except subprocess.CalledProcessError:
            # Java command not found in PATH, check JDK folder in home directory
            jdk_home = os.path.join(home_dir, "jdk", "jdk-11.0.18+10")  # Change this to your JDK installation directory
            print("Check jdk_home", jdk_home)
            java_cmd = os.path.join(jdk_home, "bin", "java.exe")
            if not os.path.exists(java_cmd):
                print("Java command not found.")
                safe_exit(1)

    with get_spinner(text=f'Starting Geowaever...', spinner='dots'):
        geoweaver_jar = os.path.join(home_dir, "geoweaver.jar")
        print(f'"{java_cmd}" -jar "{geoweaver_jar}"')
        subprocess.Popen([java_cmd, "-jar", geoweaver_jar], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, creationflags=subprocess.CREATE_NEW_CONSOLE)

        status = 0
        counter = 0
        max_attempts = 20
        retry_delay = 2
        while counter < max_attempts:
            time.sleep(retry_delay)
            counter += 1
            try:
                response = requests.get(GEOWEAVER_DEFAULT_ENDPOINT_URL, allow_redirects=False)
                if response.status_code == 302:
                    log_file = get_log_file_path()
                    # Now you can safely open the log file for reading
                    with open(log_file, "r") as f:
                        print(f.read())
                    print("Success: Geoweaver is up")
                    if exit_on_finish:
                        safe_exit(0)
            except Exception as e:
                # print(f"Error occurred during request: {e}")
                continue

        print("Error: Geoweaver is not up")
        if exit_on_finish:
            safe_exit(1)


def stop_on_windows():
    print("Stopping Geoweaver...")
    subprocess.run(["taskkill", "/f", "/im", "java.exe"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    print("Geoweaver stopped successfully.")


def check_java_exists():
    with get_spinner(text=f'Check if Java is installed...', spinner='dots'):
        specified_path = os.path.expanduser("~/jdk/jdk-11.0.18+10/bin/java")
        if os.path.isfile(specified_path):
            print(f"Using Java in home directory: {specified_path}")
            return specified_path

        # Check if default Java exists
        try:
            result = subprocess.run(["java", "-version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if result.returncode == 0:
                print("Using Java from system..")
                return "java"
        except FileNotFoundError:
            pass

        return None


def start_on_mac_linux(force_restart: bool=False, force_download: bool=False, exit_on_finish: bool=False):
    if force_restart:
        # First stop any existing Geoweaver
        stop_on_mac_linux(exit_on_finish=exit_on_finish)

    # Checking Java
    java_path = check_java_exists()
    if java_path is None:
        print("Java not found. Exiting...")
        if exit_on_finish:
            safe_exit(1)

    with get_spinner(text=f'Starting Geoweaver...', spinner='dots'):
        # Start Geoweaver
        cmds = [java_path, "-jar", os.path.expanduser("~/geoweaver.jar")]
        logger.info("Running ", " ".join(cmds))
        with open(os.path.expanduser("~/geoweaver.log"), 'w') as log_file:
            subprocess.Popen(cmds, 
                            stdout=log_file, 
                            stderr=subprocess.STDOUT)

        # Wait for Geoweaver to start
        time.sleep(2)  # Adjust as necessary

        status = 0
        counter = 0
        max_counter = 10
        while counter != max_counter:  # max wait for 20 seconds
            try:
                status = requests.get(GEOWEAVER_DEFAULT_ENDPOINT_URL).status_code
                logger.debug(f"Received code {status}")
                if status == 302 or status == 200:
                    break
            except requests.exceptions.ConnectionError:
                pass  # Connection error, retrying
            time.sleep(2)
            counter += 1

        if counter == max_counter:
            print("Error: Geoweaver is not up")
            if exit_on_finish:
                safe_exit(1)
        else:
            print("Success: Geoweaver is up")
            if exit_on_finish:
                safe_exit(0)


def find_geoweaver_processes(current_uid):
    """
    Find all Geoweaver-related processes started by the current user.
    
    :param current_uid: The UID of the current user.
    :return: A list of matching process objects.
    """
    processes = []
    for proc in psutil.process_iter(['pid', 'name', 'cmdline', 'uids']):
        try:
            if (proc.info and proc.info['cmdline'] and
                ('geoweaver.jar' in " ".join(proc.info['cmdline']) or
                 'GeoweaverApplication' in " ".join(proc.info['cmdline'])) and
                proc.info['uids'] and proc.info['uids'].real == current_uid):
                processes.append(proc)
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            continue  # Skip processes that have exited or are inaccessible
    return processes

def stop_on_mac_linux(exit_on_finish: bool = False) -> int:
    with get_spinner(text='Stopping Geoweaver...', spinner='dots'):
        logger.info("Stopping any running Geoweaver processes...")

        # Get current user's UID
        current_uid = os.getuid()

        # Find all processes running geoweaver.jar or GeoweaverApplication that are started by the current user
        processes = find_geoweaver_processes(current_uid)

        if not processes:
            print("No running Geoweaver processes found for the current user.")
            return 0

        # Attempt to kill each process
        errors = []
        for proc in processes:
            try:
                proc.terminate()
                proc.wait(timeout=5)  # Wait for the process to terminate
                logger.info(f"Successfully stopped process {proc.info['pid']}.")
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                logger.error(f"Process {proc.info['pid']} has already exited or is inaccessible.")
                errors.append(f"Process {proc.info['pid']} is not accessible.")
            except psutil.TimeoutExpired:
                logger.warning(f"Process {proc.info['pid']} did not terminate in time, forcing kill.")
                proc.kill()  # Forcefully kill if it didn't terminate in time
                errors.append(f"Process {proc.info['pid']} was forcefully killed.")

        # Log errors if any
        if errors:
            for error in errors:
                logger.error(error)
            print("Some processes could not be stopped.")
            return 1

        # Check status of Geoweaver
        status = subprocess.run(
            ["curl", "-s", "-o", "/dev/null", "-w", "%{http_code}\n", GEOWEAVER_DEFAULT_ENDPOINT_URL],
            capture_output=True, text=True
        ).stdout.strip()

        logger.info("Geoweaver status: " + status)
        if status != "302":
            print("Stopped Geoweaver successfully.")
            return 0
        else:
            print("Error: Unable to stop Geoweaver.")
            return 1


def start(force_download=False, force_restart=False, exit_on_finish=True):
    download_geoweaver_jar(overwrite=force_download)
    check_java()

    if check_os() == 3:
        logger.debug(f"Detected Windows, running start python script..")
        start_on_windows(force_restart=force_restart, force_download=force_download, exit_on_finish=exit_on_finish)
    else:
        logger.debug(f"Detected Linux/MacOs, running start python script..")
        start_on_mac_linux(force_restart=force_restart, force_download=force_download, exit_on_finish=exit_on_finish)


def stop(exit_on_finish: bool=False):
    check_java()
    if check_os() == 3:
        stop_on_windows()
    else:
        exit_code = stop_on_mac_linux()
        if exit_on_finish:
            safe_exit(exit_code)


def show(geoweaver_url=GEOWEAVER_DEFAULT_ENDPOINT_URL):
    download_geoweaver_jar()  # check if geoweaver is initialized
    check_java()
    if check_ipython():
        logger.info("enter ipython block")
        from IPython.display import IFrame

        logger.warning("This only works when the Jupyter is visited from localhost!")
        return IFrame(src=geoweaver_url, width="100%", height="500px")
    else:
        logger.info("enter self opening block")
        webbrowser.open(geoweaver_url)


def ensure_geoweaver_started():
    geoweaver_running = check_geoweaver_status()
    if not geoweaver_running:
        start(exit_on_finish=False)



from pydantic import BaseModel
from typing import Optional

class Workflow(BaseModel):
    id: str
    name: Optional[str]
    description: Optional[str]
    owner: Optional[str]
    confidential: Optional[str]
    edges: Optional[str]
    nodes: Optional[str]



from pydantic import BaseModel
from typing import Optional
import uuid
from datetime import datetime

class Checkpoint(BaseModel):
    id: Optional[uuid.UUID]
    executionId: str
    edges: str
    nodes: str
    workflow: Optional[str]
    createdAt: Optional[datetime]



import logging
import subprocess
import requests
from pygeoweaver.constants import *

from pygeoweaver.server import ensure_geoweaver_started
from pygeoweaver.utils import (
    download_geoweaver_jar,
    get_geoweaver_jar_path,
    get_java_bin_path,
    get_root_dir,
    get_spinner,
    check_os,
)
from pygeoweaver.server import (
    start_on_windows,
    start_on_mac_linux,
    check_geoweaver_status,
)

logger = logging.getLogger(__name__)

def ensure_server_running(force_restart=False, force_download=False):
    """
    Ensure the Geoweaver server is running. If not, start the sever.
    :param force_restart: Restart the server even if it's already running.
    :param force_download: Force download of Geoweaver jar if necessary.
    """
    logger.info("Ensuring Geoweaver server is running...")
    if force_download:
        download_geoweaver_jar()

    if not force_restart and check_geoweaver_status():
        logger.info("Geoweaver server is already running.")
        return

    logger.info("Starting Geoweaver server...")
    os_type = check_os()
    try:
        if os_type == 3:  # Windows
            start_on_windows(force_restart=force_restart, exit_on_finish=False)
        else:  # macOS or Linux
            start_on_mac_linux(force_restart=force_restart, exit_on_finish=False)

        # Recheck server status
        if not check_geoweaver_status():
            raise RuntimeError("Failed to start Geoweaver server.")
        logger.info("Geoweaver server started successfully.")
    except Exception as e:
        logger.error(f"Error starting Geoweaver server: {e}")
        raise


def get_geoweaver_endpoint():
    """
    Get the endpoint URL for the Geoweaver server.
    """
    return GEOWEAVER_DEFAULT_ENDPOINT_URL or "http://localhost:8070/Geoweaver"


def detail_workflow(workflow_id, force_restart=False, force_download=False):
    """
    Display detailed information about a workflow.
    :param workflow_id: The ID of the workflow.
    :param force_restart: Restart the server even if it's already running.
    :param force_download: Force download of Geoweaver jar if necessary.
    """
    if not workflow_id:
        raise RuntimeError("Workflow id is missing")
    with get_spinner(text="Getting workflow details..", spinner="dots"):
        download_geoweaver_jar()
        ensure_geoweaver_started()
        process = subprocess.run(
            [
                get_java_bin_path(),
                "-jar",
                get_geoweaver_jar_path(),
                "detail",
                f"--workflow-id={workflow_id}",
            ],
            cwd=f"{get_root_dir()}/",
        )
    
    print("\n", process.stdout)
    if process.stderr:
        print("=== Error ===")
        print(process.stderr)
        logger.error(process.stderr)
        

def detail_process(process_id):
    """
    Display detailed information about a process.
    :param process_id: The ID of the process.
    :param force_restart: Restart the server even if it's already running.
    :param force_download: Force download of Geoweaver jar if necessary.
    """
    if not process_id:
        raise RuntimeError("Process id is missing")
    with get_spinner(text="Getting process details..", spinner="dots"):
        download_geoweaver_jar()
        ensure_geoweaver_started()
        process = subprocess.run(
            [
                "java", "-jar", get_geoweaver_jar_path(), "detail",
                "--process-id", process_id
            ],
            text=True,
            capture_output=True
        )
    print("\n", process.stdout)
    if process.stderr:
        print("=== Error ===")
        print(process.stderr)
        logger.error(process.stderr)

        if process.stderr:
            print("=== Error ===")
            print(process.stderr)
            logger.error(process.stderr)
        if process.stdout:
            print("=== Process Details ===")
            print(process.stdout)


def detail_host(host_id, force_restart=False, force_download=False):
    """
    Display detailed information about a host.
    :param host_id: The ID of the host.
    :param force_restart: Restart the server even if it's already running.
    :param force_download: Force download of Geoweaver jar if necessary.
    """
    if not host_id:
        raise RuntimeError("Host id is missing")
    with get_spinner(text="Getting host details..\n", spinner="dots"):
        download_geoweaver_jar()
        ensure_geoweaver_started()
        process = subprocess.run(
            [
                "java", "-jar", get_geoweaver_jar_path(), "detail",
                "--host-id", host_id
            ],
            capture_output=True,
            text=True
        )
    
        print("\n" + process.stdout)
        if process.stderr:
            print("=== Error ===")
            print(process.stderr)
            logger.error(process.stderr)

def get_process_code(process_id):
    """
    Get the code of a process.

    :param process_id: The ID of the process.
    :type process_id: str
    :return: The code of the process.
    :rtype: str
    """
    ensure_geoweaver_started()
    r = requests.post(
        f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/detail",
        data={"type": "process", "id": process_id},
    ).json()
    return r["code"]



from pydantic import BaseModel, Field
from typing import Optional

class Environment(BaseModel):
    id: str
    name: str
    type: str
    bin: str
    pyenv: str
    basedir: str
    hostid: Optional[str] = None  # Use Optional to allow for null values
    settings: Optional[str] = None  # Use Optional to allow for null values



"""Module for upgrading Geoweaver by downloading the latest JAR file."""

import logging
import os

from pygeoweaver.server import stop, start, check_geoweaver_status
from pygeoweaver.utils import get_spinner, get_geoweaver_jar_path, download_geoweaver_jar
from pygeoweaver.jdk_utils import check_java

logger = logging.getLogger(__name__)

def upgrade_geoweaver(force=False, no_start=False):
    """
    Upgrade Geoweaver by downloading the latest JAR file.
    
    This function follows these steps:
    1. Ask for confirmation as the process will stop Geoweaver if running (unless force=True)
    2. Stop Geoweaver if it's running
    3. Force download the latest Geoweaver JAR file
    4. Start Geoweaver (unless no_start=True)
    
    Args:
        force (bool): If True, skip the confirmation prompt
        no_start (bool): If True, don't start Geoweaver after upgrade
    
    Returns:
        bool: True if the operation was successful, False otherwise.
    """
    # Check if Java is installed
    check_java()
    
    # Ask for confirmation before proceeding (unless force=True)
    if not force:
        print("\nWARNING: This upgrade will stop Geoweaver if it's currently running.")
        confirmation = input("Are you sure you want to upgrade Geoweaver right now? (yes/no): ")
        
        if confirmation.lower() != 'yes':
            print("Upgrade cancelled.")
            return False
    
    # Step 1: Stop Geoweaver if it's running
    with get_spinner(text="Checking if Geoweaver is running...", spinner="dots"):
        if check_geoweaver_status():
            logger.info("Stopping Geoweaver...")
            stop(exit_on_finish=False)
        else:
            logger.info("Geoweaver is not running.")
    
    # Step 2: Force download the latest Geoweaver JAR file
    jar_path = get_geoweaver_jar_path()
    if os.path.exists(jar_path):
        with get_spinner(text="Removing existing Geoweaver JAR file...", spinner="dots"):
            try:
                os.remove(jar_path)
                logger.info(f"Removed existing JAR file at {jar_path}")
            except Exception as e:
                logger.error(f"Failed to remove existing JAR file: {str(e)}")
                return False
    
    with get_spinner(text="Downloading latest Geoweaver JAR file...", spinner="dots"):
        try:
            download_geoweaver_jar(overwrite=True)
            logger.info("Successfully downloaded the latest Geoweaver JAR file")
        except Exception as e:
            logger.error(f"Failed to download Geoweaver JAR file: {str(e)}")
            return False
    
    # Step 3: Start Geoweaver (unless no_start=True)
    if not no_start:
        with get_spinner(text="Starting Geoweaver...", spinner="dots"):
            start(exit_on_finish=False)
    else:
        print("\nSkipping Geoweaver startup as requested.")
    
    print("\nGeoweaver upgrade completed successfully!")
    print(f"The latest Geoweaver JAR file has been downloaded to {jar_path}")
    
    return True


import time
import itertools
import threading
from IPython.display import display, clear_output

spinner_styles = {
    "dots": itertools.cycle(['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏']),
    "line": itertools.cycle(['-', '/', '|', '\\']),
    "arrow": itertools.cycle(['←', '↖', '↑', '↗', '→', '↘', '↓', '↙']),
    "circle": itertools.cycle(['◐', '◓', '◑', '◒']),
    "bounce": itertools.cycle(['⠁', '⠂', '⠄', '⡀', '⢀', '⠠', '⠐', '⠈'])
}

class Spinner:
    def __init__(self, text="Loading...", style="circle", interval=0.1):
        self.text = text
        self.interval = interval
        self.spinner_cycle = spinner_styles.get(style, spinner_styles["dots"])
        self.stop_running = threading.Event()

    def start(self):
        threading.Thread(target=self.run, daemon=True).start()

    def run(self):
        while not self.stop_running.is_set():
            display(f"{next(self.spinner_cycle)} {self.text}", display_id='spinner')
            time.sleep(self.interval)
            clear_output(wait=True)

    def stop(self):
        self.stop_running.set()
        display(f"{self.text} Done!", display_id='spinner')

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()
        



import os

GEOWEAVER_PORT=os.getenv("GEOWEAVER_PORT", "8070")
GEOWEAVER_DEFAULT_ENDPOINT_URL = f"http://localhost:{GEOWEAVER_PORT}/Geoweaver"
COMMON_API_HEADER = {"Content-Type": "application/json"}
GEOWEAVER_URL = (
    "https://github.com/ESIPFed/Geoweaver/releases/download/latest/geoweaver.jar"
)




from pygeoweaver.api_call.pgw_base_api_caller import BaseAPI


class GeoweaverWorkflowAPI(BaseAPI):
    def add_workflow(self, workflow_data):
        return self._call_api('/add/workflow', method='POST', data=workflow_data)

    def edit_workflow(self, workflow_data):
        return self._call_api('/edit/workflow', method='POST', data=workflow_data)

    def get_workflow(self, workflow_id):
        return self._call_api(f'/get/workflow/{workflow_id}', method='GET')

    def delete_workflow(self, workflow_id):
        return self._call_api(f'/delete/workflow/{workflow_id}', method='DELETE')




from pydantic import BaseModel, Field
from typing import Optional, Set

class Host(BaseModel):
    id: str
    name: str
    ip: str
    port: str
    username: str
    owner: str
    type: str
    url: str
    confidential: bool
    envs: Optional[Set[str]] = None




import requests
from pygeoweaver.constants import *
import pandas as pd

from pygeoweaver.utils import check_ipython


def get_process_by_name(process_name):
    """
    Get processes by their name.

    :param process_name: The name of the process.
    :type process_name: str
    """
    response = requests.post(
        f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/list", data={"type": "process"}
    )
    process_list = response.json()

    matching_processes = []

    for process in process_list:
        if process["name"] == process_name:
            matching_processes.append(process)
    pd.set_option("display.max_columns", None)  # Display all columns
    pd.set_option("display.max_rows", None)  # Display all rows
    pd.set_option("display.expand_frame_repr", False)  # Prevent truncation of columns
    pd.DataFrame(matching_processes)
    if not check_ipython():
        return pd.DataFrame(matching_processes)


def get_process_by_id(process_id):
    """
    Get a process by its ID.

    :param process_id: The ID of the process.
    :type process_id: str
    """
    response = requests.post(
        f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/list", data={"type": "process"}
    )
    process_list = response.json()

    matching_processes = []

    for process in process_list:
        if process["id"] == process_id:
            matching_processes.append(process)
    pd.set_option("display.max_columns", None)  # Display all columns
    pd.set_option("display.max_rows", None)  # Display all rows
    pd.set_option("display.expand_frame_repr", False)  # Prevent truncation of columns
    pd.DataFrame(matching_processes)
    if not check_ipython():
        return pd.DataFrame(matching_processes)


def get_process_by_language(language):
    """
    Get processes by their programming language.

    :param language: The programming language of the processes.
    :type language: str
    """
    response = requests.post(
        f"{GEOWEAVER_DEFAULT_ENDPOINT_URL}/web/list", data={"type": "process"}
    )
    process_list = response.json()

    matching_processes = []

    for process in process_list:
        if process["lang"] == language:
            matching_processes.append(process)
    pd.set_option("display.max_columns", None)  # Display all columns
    pd.set_option("display.max_rows", None)  # Display all rows
    pd.set_option("display.expand_frame_repr", False)  # Prevent truncation of columns
    pd.DataFrame(matching_processes)
    if not check_ipython():
        return pd.DataFrame(matching_processes)
